"""Read EMA data stored in an Excel 2007- (xlsx) workbook file.

Each Excel work-book file, and each work-sheet in the file,
may include EMA records for
one or more subject(s), each reporting on one or more EMA variables,
representing nominal SCENARIO categories and/or ordinal ATTRIBUTE ratings.

Each row in the worksheet must include enough data to identify
all requested variables.

*** Usage Example:
    See module ema_data and main template script run_ema

*** Version History:
* Version 0.5.1:
2021-11-26, handle missing file field nicely

* Version 0.5:
2021-10-20, first functional version
2021-11-17, save method
"""
# ********** check file exceptions **************
from openpyxl import load_workbook, Workbook
from openpyxl.utils.exceptions import InvalidFileException
from zipfile import BadZipFile
# because Openpyxl might try to open bad xlsx file as a zip archive
from openpyxl.utils import column_index_from_string
import logging

from EmaCalc import ema_file

logger = logging.getLogger(__name__)


class FileFormatError(ema_file.FileReadError):
    """Format error causing non-usable data"""


class ParameterError(ema_file.FileReadError):
    """Error in calling parameters causing non-usable data in some file."""


class ArgumentError(RuntimeError):
    """Error in calling arguments. No file can be read."""


class EmaFile(ema_file.EmaFile):
    """Interface to Excel xlsx workbook file storing count-profile data.
    Each sheet in the workbook may include data for
    one or more subjects, count factors.
    Data elements MUST be stored at the same location in all sheets.
    """
    def __init__(self, file_path,
                 ema_vars,
                 top_row=2,
                 sheets=None,
                 subject=None,
                 ):
        """File interface to data stored in an excel file,
        decoded as a sequence of PairedCompItem instances.
        :param file_path: Path to file for reading
        :param top_row: integer address of first row containing PairedCompItem data
        :param sheets: (optional) list of sheet names to be searched for data.
        :param subject: 'sheet' or column for subject identification label
            **** allow subject = 'file' name ? *******
        :param ema_vars: dict with elements (ema_key, location), where
            ema_key is a string identifying one count key,
            location is a string with a column (top) address like 'D'.
        """
        super().__init__(file_path)
        self.subject = _check_column_or_sheet(subject)
        self.ema_vars = ema_vars
        self.sheets = sheets
        self.top_row = top_row
        # self.test_factors = _check_test_cond(test_factors)

    def __iter__(self):
        """Generator yielding data from an excel file.
        :return: generator yielding EmaRecord instances
        """
        try:
            wb = load_workbook(str(self.file_path), read_only=True)
        except Exception as e:  # openpyxl might raise InvalidFileException, BadZipFile, other?
            raise FileFormatError(f'Cannot load Excel workbook from file {self.file_path.stem}. '
                                  + f'Openpyxl error: {e}')
        if self.sheets is None:
            sheets = wb.sheetnames
        else:
            sheets = set(self.sheets) & set(wb.sheetnames)
        if len(sheets) == 0:
            raise FileFormatError(f'No accepted sheets found in {self.file_path.stem}')
        for sheet_name in sheets:
            ws = wb[sheet_name]
            rows = ws.rows
            for _ in range(self.top_row - 1):
                try:
                    row = rows.__next__()
                    logger.debug(f'skipping row {row[0].row}')
                except StopIteration:
                    raise FileFormatError(f'No data rows in {self.file_path.stem}.'
                                          + f'sheet {repr(sheet_name)}')
            for row in rows:
                r = ema_file.EmaRecord(subject=self._get_subject(ws, row),
                                       ema=self._get_ema_dict(ws, row)
                                       )
                if r.subject is not None:
                    # or any(r_i is None  # **** let caller check this!
                    #      for r_i in r.ema.values())):
                    # we have a valid record
                    yield r  # ******
                else:
                    logger.debug(f'not using row {row[0]}...')

    def _get_subject(self, ws, row):
        """
        :param ws: a worksheet
        :param row: tuple of openpyxl Cell instances
        :return: subject code, integer or string
        """
        return _get_value(ws, self.subject, row)

    def _get_ema_dict(self, ws, row):
        """
        :param ws: a worksheet
        :param row: tuple of openpyxl Cell instances
        :return: dict with tuples (ema_key, category_label), where
            ema_key is string defining an EMA 'dimension'
            category_label = numeric or string label of response category in this factor
        """
        return dict([(cf, _get_value(ws, c_col, row))
                     for (cf, c_col) in self.ema_vars.items()])

    # ----------------------------------------------- save data to workbook
    def save(self, subject, items, allow_over_write=False):
        """Save EMA records to file with ONE sheet for ONE subject,
        identified in the sheet title.
        :param subject: string with subject id
        :param items: iterable of dicts with EMA data
        :param allow_over_write: boolean switch, =True allows old file to be over-written
        :return: None

        Result: a new file is created.

        2021-11-17, new method for this file format
        """
        # *** update self properties to allow reading from the file?
        wb = Workbook(write_only=True)
        ws = wb.create_sheet(title=subject)
        header = self._make_header()
        n_col = len(header)
        ws.append(header)
        self.top_row = 2
        for ema in items:
            ws.append(self._make_row(ema, n_col))
        if not allow_over_write:
            self.file_path = ema_file.safe_file_path(self.file_path)
        self.sheets = wb.sheetnames
        wb.save(self.file_path)

    def _make_header(self):  # **** module function or static ******
        """Make header list from column addresses of self.
        :param n_col:
        :return: list of header labels, length = n_col
        """
        n_col = max(column_index_from_string(v)
                    for v in self.ema_vars.values())
        h = [None for _ in range(n_col)]
        for (k, c) in self.ema_vars.items():
            h[column_index_from_string(c) - 1] = k
        return h

    def _make_row(self, ema, n_col):
        """Make list with contents from given ema record
        :param ema: a dict with EMA data
        :return: r = list filled with elements from ema,
            placed at columns defined by self.ema_vars
        """
        r = [None for _ in range(n_col)]
        for (k, v) in ema.items():
            if k in self.ema_vars.keys():
                r[column_index_from_string(self.ema_vars[k]) - 1] = v
        return r


# --------------------------------------------------- help sub-functions

def _check_column(col):
    """Check that a parameter address is acceptable
    """
    if col is None or type(col) is str and col.isupper():
        return col
    # otherwise:
    raise ArgumentError(f'Column address {col} must be string of uppercase letters')


def _check_column_or_sheet(col):
    """Check that a parameter address is either 'sheet' or a column string address
    """
    if type(col) is str and col == 'sheet':
        return col
    else:
        return _check_column(col)


def _get_value(ws, col, row):
    """Get contents in ONE cell or in sheet title
    :param ws: a worksheet
    :param col: one column address or 'sheet'
    :param row: integer address of row
    :return: cell contents
    """
    if col == 'sheet':
        return ws.title
    # c = ws[col + str(row)]
    try:
        cell = row[column_index_from_string(col) - 1]
        return cell.value
    except IndexError:
        return None  # OR raise ArgumentError ?




"""General utility class and functions for internal EmaCalc use.

Defines class
EmaParamBase --- container for common properties used by all model classes
    specifying the indexing of separate types of parameters.

NOTE: To make the model identifiable,
although both latent-variable locations and response thresholds are freely variable,
some restriction is necessary.

This behavior is user-controlled by initialization arguments
restrict_attribute and restrict_threshold.

The model must also be slightly restricted by a weakly informative prior,
for numerical stability in case of extreme response patterns,
e.g., ALL ratings in the highest ordinal category.


*** Version history:
* Version 0.8.1
2022-02-28, changed class name -> EmaParamBase
2022-02-23, GMM comp moved to ema_model.EmaGroupModel, separate GMM for each group

* Version 0.7.1
2022-01-11, _make_prior with log normalized uniform scenario prob

* Version 0.7
2022-01-02, minor cleanup _make_prior: ONLY scalar precision shape for GMM components

* Version 0.6
2021-12-03, user-definable prior inter-individual variance for reference parameters
2021-12-05, testing ways to set hyper-prior
2021-12-06, boolean properties restrict_attribute, restrict_threshold of EmaParamBase
            for user-specified model restriction
2021-12-08, minor checks against numerical overflow

* Version 0.5.1
2021-11-26, allow ZERO Attributes, i.e., empty emf.attribute_grades
2021-12-01, cleanup some doc comments
2021-12-02, Attribute sensory location fixed == 0. for first Scenario category
            regardless of regression_effect specification.

* Version 0.5
2021-11-24, first published beta version
"""
import numpy as np
from scipy.special import logit, expit
import logging

from EmaCalc.gauss_gamma import GaussianRV

# ------------------------------------------------------
__version__ = "2022-02-28"

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST

PRIOR_PSEUDO_RESPONDENT = 0.1
# = hyperprior total pseudo-count re ONE real respondent
# = prior GaussianRV.mean.learned_weight for all GMM components.
# This value is NOT critical for Occam's Razor behavior.

PRIOR_PARAM_SCALE = 1.  # 0.5?
# = main hyperprior scale of most Gaussian model parameters,
# with attribute parameters rescaled with the class scale of latent sensory variable.
# This prior may have effect on the Occam's Razor function:
# small values may allow several active GMM components with small variance,
# but experiments suggest the prior value is not critical.

PRIOR_PREC_A = PRIOR_PSEUDO_RESPONDENT / 2
# = GaussianRV precision shape for ALL parameters

PRIOR_PREC_B = PRIOR_PARAM_SCALE**2 / 2
# = GaussianRV precision inv_scale for MOST EmaModel parameters
# -> allows very small precision <=> large inter-individual variance
# -> mode of prior component-element variance = PRIOR_PREC_B /(PRIOR_PREC_A + 1) approx= PRIOR_PREC_B


# ------------------------------------------------------------------
class EmaParamBase:
    """Container for common properties of ema_model.EmaModel and its sub-models.

    Each individual parameter distribution is represented by
    a large set of samples, stored as property EmaSubjectModel.xi, with
    xi[s, :] = s-th sample of the parameter vector for ONE subject.

    All model classes share mapping properties defined here,
    for extracting parameter subsets
    from the array of parameter vectors.
    """
    def __init__(self, emf, effects, rv_class,
                 theta_map,
                 scenario_slices,
                 attribute_slices,
                 comp_prior,
                 restrict_attribute=False,
                 restrict_threshold=False):
        """
        :param emf: single ema_data.EmaFrame instance,
        :param effects: iterable with regression-effect terms for attribute regression
            effects[i] = single key in emf.scenarios.keys() or tuple of such keys
        :param rv_class: latent sensory variable class,
            defining its distribution as either logistic or normal.
        :param theta_map: fixed 2D array to extract latent-variable location samples from xi array
        :param scenario_slices: list of slice objects, such that
            xi[:, scenario_slices[t]] = alpha = log-prob for scenarios in t-th test stage
        :param attribute_slices: list of slice objects, such that
            xi[:, attribute_slices[i]] = all (beta, eta) parameters for i-th Attribute
            where
            beta = xi[:, attribute_slices[i]][:, :n_beta] = regression-effect parameters
            eta = xi[:, attribute_slices[i]][:, n_beta:] = threshold parameters
            where
            n_beta = n of regression-effect parameters, same for each attribute question.
        :param comp_prior: single GaussianRV instance, prior for ALL GMM components,
        :param restrict_attribute: (optional) boolean switch
            to force restriction on attribute sensory-variable locations
        :param restrict_threshold: (optional) boolean switch
            to force restriction on response-threshold locations
        """
        self.emf = emf
        self.effects = effects
        self.rv_class = rv_class
        self.theta_map = theta_map
        self.scenario_slices = scenario_slices
        self.attribute_slices = attribute_slices
        self.comp_prior = comp_prior
        self.restrict_attribute = restrict_attribute
        self.restrict_threshold = restrict_threshold

    @classmethod
    def initialize(cls, emf, effects, rv_class,
                   restrict_attribute=False,
                   restrict_threshold=False):
        """Assign all parameter-extraction properties, and
        GaussianRV mixture components with correct size.
        :param emf: single ema_data.EmaFrame instance,
            defining the experiment structure.
        :param effects: iterable with regression-effect terms for attribute regression
            effects[i] = single key in emf.scenarios.keys() or tuple of such keys
        :param rv_class: latent sensory variable class,
            defining its distribution as either logistic or normal.
        :param restrict_attribute: (optional) boolean switch
            to force restriction on attribute sensory-variable locations
        :param restrict_threshold: (optional) boolean switch
            to force restriction on response-threshold locations
        :return: None

        Result: all properties initialized
        """
        _check_effects(emf, effects)
        effects = [e_i if type(e_i) is tuple else (e_i,)
                   for e_i in effects]
        # = list of requested regression effects of scenario categories
        # to be estimated for their influence on ATTRIBUTE responses.
        # Each element MUST be a tuple with one or more scenario key(s).

        # Define mapping indices from rows of xi array to its parts:
        n_stages = emf.scenario_shape[0]
        n_scenarios = np.prod(emf.scenario_shape[1:], dtype=int)  # except stages
        scenario_slices = _make_slices(n_stages * [n_scenarios])
        # = list of slice objects, such that
        # xi[:, cls.scenario_slices[t]] = alpha = log-prob for scenarios in t-th test stage
        # i.e., n_stages slices with equal lengths, for all test stages.

        theta_map = _theta_map(emf, effects)  # , restrict_attribute)
        # = fixed 2D array to extract latent-variable location samples from xi array

        n_beta = theta_map.shape[0]
        # = number of regression-effect parameters, same for every Attribute
        n_attr_param = [n_beta + len(ordinal_levels)
                        for ordinal_levels in emf.attribute_grades.values()]
        # n_attr_param[i] = total number of model params for i-th Attribute
        # NOTE: May be EMPTY if no Attributes are defined
        attribute_slices = _make_slices(n_attr_param,
                                        start=scenario_slices[-1].stop)
        # xi[:, cls.attribute_slices[i]] = all (beta, eta) parameters for i-th Attribute
        # where
        # beta = xi[:, cls.attribute_slices[i]][:, :n_beta] = regression-effect parameters
        # eta = xi[:, cls.attribute_slices[i]][:, n_beta:] = threshold parameters
        # where
        # n_beta = n of regression-effect parameters, same for each attribute question.
        if len(attribute_slices) > 0:
            n_parameters = attribute_slices[-1].stop
        else:
            n_parameters = scenario_slices[-1].stop

        # define mixture components:
        comp_prior = _make_prior(n_parameters, n_beta,
                                 scenario_slices,
                                 attribute_slices, rv_class)
        # = single GaussianRV instance, prior for ALL GMM components,
        # fixed throughout the VI learning procedure.
        logger.debug(f'PRIOR_PSEUDO_RESPONDENT = {PRIOR_PSEUDO_RESPONDENT:.3f}; ' +
                     f'PRIOR_PARAM_SCALE = {PRIOR_PARAM_SCALE:.3f}; ' +
                     f'PRIOR_PREC_A = {PRIOR_PREC_A:.3f}; ' +
                     f'PRIOR_PREC_B = {PRIOR_PREC_B:.3f}')
        logger.debug(f'comp_prior.prec.a = {comp_prior.prec.a:.3f}')
        logger.debug('comp_prior.prec.b = '
                     + np.array_str(comp_prior.prec.b,
                                    precision=5))
        logger.debug('mode{1 / comp_prior.prec}= '
                     + np.array_str(comp_prior.prec.mode_inv(),
                                    precision=5))
        return cls(emf, effects, rv_class,
                   theta_map,
                   scenario_slices,
                   attribute_slices,
                   comp_prior,
                   restrict_attribute,
                   restrict_threshold)

    @property
    def n_parameters(self):
        if len(self.attribute_slices) > 0:
            return self.attribute_slices[-1].stop
        else:
            return self.scenario_slices[-1].stop

    @property
    def n_beta(self):
        """Total number of regression-effect parameters, same for each Attribute
        """
        return self.theta_map.shape[0]

    @property
    def attribute_slice_dict(self):
        """
        :return: dict with elements (attribute_key, attribute_slice)
        """
        return dict((a, a_slice)
                    for (a, a_slice) in zip(self.emf.attribute_grades.keys(),
                                            self.attribute_slices))

    def beta_0_size(self):
        """number of effect parameters for FIRST effect term
        :return: n_beta = scalar

        NOTE: matches sizes defined in _theta_map()
        """
        return np.prod([len(self.emf.scenarios[sc_i]) for sc_i in self.effects[0]],
                       dtype=int)

    def scenario_prob(self, xi):
        """Extract probability-mass for scenarios, given parameters,
        used mainly by ema_display
        :param xi: 2D array with parameter samples
        :return: u = mD array with scenario probability-mass within each Stage,
            u[s, k0, k1, k2, ...] = s-th sample of P[(k1, k2,...)-th scenario | stage k0]
            sum u[s, k0] == 1., for all s and k0

        2022-02-15, check for underflow
        """
        n_sc = self.scenario_slices[-1].stop
        alpha = xi[:, :n_sc].reshape((xi.shape[0], self.emf.n_stages, -1))
        alpha -= np.amax(alpha, axis=-1, keepdims=True)
        too_small = np.all(alpha < np.log(np.finfo(float).tiny), axis=-1)
        n_underflow = np.sum(too_small)
        if n_underflow > 0:
            logger.warning(f'scenario_prob: {n_underflow} alpha samples too small. '
                           + 'Should not happen. Maybe too few responses?')
            logger.debug(f'alpha[too_small, :] = {alpha[too_small]}')
        u = np.exp(alpha)
        # avoid nan after normalization, should not be needed !
        u /= np.sum(u, axis=-1, keepdims=True)
        return u.reshape((-1, *self.emf.scenario_shape))

    def attribute_theta(self, xi, a):
        """Extract location of latent sensory variable, for given attribute
        :param xi: 2D array with parameter sample vectors
            xi[s, :] = s-th parameter sample vector
        :param a: attribute key = one of self.emf.attribute_grades.keys()
        :return: theta = mD array, with
            theta[s, k0, k1, ...] = s-th sample of
                attribute location, given the (k0, k1, ...)-th scenario.
        """
        a_slice = self.attribute_slice_dict[a]
        beta = xi[..., a_slice][..., :self.n_beta]
        return np.dot(beta, self.theta_map).reshape((-1, *self.emf.scenario_shape))

    def attribute_tau(self, xi, a):
        """Extract response thresholds for given attribute
        :param xi: 2D array with parameter sample vectors
            xi[s, :] = s-th parameter sample vector
        :param a: attribute key = one of self.emf.attribute_grades.keys()
        :return: tau = mD array, with
            tau[s, l] = s-th sample of UPPER limit of l-th response interval,
                EXCEPT the last at +inf
                tau.shape[-1] == len(self.emf.attribute_grades[a]) - 1
        """
        a_slice = self.attribute_slice_dict[a]
        eta = xi[..., a_slice][..., self.n_beta:]
        return response_thresholds(eta)[..., 1:-1]


# ---------------------------------------- Module help functions:

def response_thresholds(log_w):
    """Transform given log-category-width parameters to response thresholds.
    :param log_w: 1D or 2D array with
        log_w[..., m] = ...-th sample of log non-normalized width of m-th interval
            in mapped domain [0, 1].
        log_w.shape[-1] == M == number of response-scale intervals.
    :return: tau = 1D or 2D array, with all elements in [-inf, +inf]
        (tau[..., m], tau[..., m+1] = (LOWER, UPPER) limits for m-th ordinal response interval
        tau[..., 0] ==  - np.inf
        tau[..., -1] == + np.inf
        tau.ndim == log_w.ndim; tau.shape[-1] == log_w.shape[-1] + 1

    Method:
        Normalized widths and interval limits are defined in transformed domain [0, 1.],
        using a logistic mapping function,
        y = expit(tau), where y in [0, 1],
            y[..., 0] = 0.
            y[..., m+1] =  (w_0 +...+ w_m) / (w_0 + ... + w_{M-1};  0 <= m <= M-1
            w_m = exp(log_w[..., m])
        Thus, thresholds are calculated with the inverse transform
        tau = logit(y)
    """
    # w = np.exp(log_w)
    # = non-normalized width of transformed intervals, always > 0
    n_overflow = np.sum(log_w > np.log(np.finfo(float).max))
    if n_overflow > 0:
        logger.warning(f'response_thresholds: {n_overflow} log_w samples too large. '
                       + 'Maybe too few responses?')
        log_w = np.minimum(log_w, np.log(np.finfo(float).max) - 1.)
        # just to avoid numerical overflow in exp(log_w), should not be needed !
    cum_w = np.cumsum(np.exp(log_w), axis=-1)
    z_shape = (*cum_w.shape[:-1], 1)
    cum_w = np.concatenate((np.zeros(z_shape), cum_w),
                           axis=-1)
    # sum_w = cum_w[..., -1:]
    return logit(cum_w / cum_w[..., -1:])


def d_response_thresholds(log_w):
    """Jacobian of cat_limits with respect to log_w
    :param log_w: 1D or 2D array (called eta in Leijon doc), with
        log_w[..., m] = ...-th sample of log non-normalized width of m-th interval,
        log_w.shape[-1] = M = number of response intervals
    :return: 2D or 3D array d_tau, with
        d_tau[..., m, i] = d tau[..., m] / d log_w[..., i]; m = 0,..., M-2; i = 0, ..., M-1
            where (tau[s, m], tau[s, m+1] = (LOWER, UPPER) limits of m-th response interval
        d_tau[..., 0, :] = d_tau[..., -1, :] = 0.; for extreme limits at +-inf
        d_tau.shape == (N, M+1, M); (N, M) = log_w

    2021-10-31, tested OK
    """
    w = np.exp(log_w)
    # (n_samples, nw) = w.shape
    nw = w.shape[-1]
    cum_w = np.cumsum(w, axis=-1)
    cw = cum_w[..., :-1, np.newaxis]  # only inner limits
    sw = cum_w[..., -1:, np.newaxis]
    # tau[..., m+1] = ln cw[..., m]  - ln (sw[..., 0] - cw[..., m])
    # dcw_dw[..., m, i] = dcw[..., m] / dw[..., i]  = 1. if i <= m else 0.
    dcw_dw = np.tril(np.ones((nw - 1, nw), dtype=int))
    dtau_dw = dcw_dw / cw - (1 - dcw_dw) / (sw - cw)
    dtau_dlogw = dtau_dw * w[..., np.newaxis, :]
    # z_shape = (n_samples, 1, nw)
    z_shape = list(dtau_dlogw.shape)
    z_shape[-2] = 1
    return np.concatenate((np.zeros(z_shape),  # *** use np.pad ???
                           dtau_dlogw,
                           np.zeros(z_shape)), axis=-2)


def tau_inv(tau):
    """Inverse of response_thresholds(log_w)
    :param tau: 1D or 2D array tau, with all elements in (-inf, +inf), EXCEPT extremes at inf
        tau[..., m] = UPPER limit for m-th interval,
            = LOWER limit for the (m+1)-th interval
        tau.shape[-1] == M - 1 == number of response intervals - 1
    :return: log_w: 1D or 2D array with
        log_w[..., m] = ...-th sample of log non-normalized width of m-th interval.
        log_w.shape[-1] == M == number of response intervals.

    Method:
        Normalized widths and interval limits are defined in transformed domain (0, 1.),
        using a logistic mapping function,
        y = expit(tau), where y in (0, 1]
            y[..., m] =  (w_0 +...+ w_m) / (w_0 + ... + w_{M-1};  0 <= m <= M-1
            w_m = exp(log_w[..., m])
    """
    y = expit(tau)
    cat_shape = (*y.shape[:-1], 1)
    y = np.concatenate((np.zeros(cat_shape),
                        y,
                        np.ones(cat_shape)), axis=-1)
    w = np.diff(y, axis=-1)
    return np.log(w)


# ---------------------------------------- private help function:
def _make_prior(n_parameters, n_beta,
                scenario_slices,
                attribute_slices, rv_class):
    """Create hyperprior for parameter distribution in the total population
    :param n_parameters: total number of parameters
    :param scenario_slices: list with index slice for each Scenario Stage
    :param attribute_slices: list with index slice for each attribute
    :param n_beta: number of regression-effect parameters for each attribute
    :return: single GaussianRV instance
    """
    prec_a = PRIOR_PREC_A  # ***** scalar, same for all elements
    prec_b = PRIOR_PREC_B * np.ones(n_parameters)
    for a_slice in attribute_slices:
        prec_b[a_slice][:n_beta] *= rv_class.scale**2
    loc = np.zeros(n_parameters)
    for sc_slice in scenario_slices:
        loc[sc_slice] = - np.log(sc_slice.stop - sc_slice.start)
    return GaussianRV.initialize(loc=loc,
                                 learned_weight=PRIOR_PSEUDO_RESPONDENT,
                                 prec_a=prec_a,
                                 prec_b=prec_b)


def _check_effects(emf, effects):
    """Check that all effects refer to unique scenario keys
    :param emf: ema_data.EmaFrame instance
    :param effects: iterable with regression-effect specifications
        effects[i] = single key in emf.scenarios.keys(), OR tuple of such keys
    :return: None
    Result: raise RuntimeError if incorrect effects
    """
    effect_keys = sum((list(e_i) if type(e_i) is tuple else [e_i]
                      for e_i in effects), start=[])
    if len(set(effect_keys)) != len(effect_keys):
        raise RuntimeError('scenario keys can occur only ONCE in regression effects')
    # *** Check no effect of scenario dimension with only ONE category
    for sc in emf.scenarios.keys():
        if len(emf.scenarios[sc]) > 1 and sc not in effect_keys:
            logger.warning(f'Scenario dim {repr(sc)} to be disregarded in regression?')
    for e_i in effect_keys:
        if e_i not in emf.scenarios.keys():
            raise RuntimeError(f'regression effect key {e_i} is not a scenario key')


def _make_slices(lengths, start=0):
    """Create a sequence of consecutive index slices, with given lengths
    :param lengths: sequence with desired slice sizes
    :param start: (optional) start index of first slice
    :return: slice_list; len(slice_list) == len(l)
    """
    slice_list = []
    for l_i in lengths:
        slice_list.append(slice(start, start + l_i))
        start += l_i
    return slice_list


def _theta_map(emf, effects):
    """Create 2D array for extraction of attribute locations
    from parameter vector samples
    :param emf: ema_data.EmaFrame instance,
        defining the analysis model structure.
    :param effects: iterable with regression-effect specifications
        effects[i] = tuple with one or more key(s) from emf.scenarios.keys()
    :return: th = 2D binary array, with
        th[j, k] = 1, IFF parameter
        beta[:, j] = xi[:, attr_slice][:, j]
        is the regression effect of k-th <=> (k0, k1, ...)-th scenario on attribute location theta,
        and xi is a 2D array of row parameter-vector samples.
    """
    def theta_one(effect):
        """Make theta_map part for ONE effect term
        :param effect: tuple of one or more scenario keys
        :return: mD array th, with
            th[j, k0, k1, ...] = j-th effect given (k0, k1, ...)-th scenario
            th[j, 0, 0, ...] = 0, in 0-th reference scenario.
            th.shape == (size_effect, *emf.scenario_shape), where
            size_effect = size of category array defined by effect.
        """
        beta_shape = tuple(len(emf.scenarios[sc_i]) for sc_i in effect)
        # = scenario_shape, for this effect term
        beta_ndim = len(beta_shape)
        # beta_size = int(np.prod(beta_shape, dtype=int))
        beta_size = np.prod(beta_shape, dtype=int)
        t = np.eye(beta_size).reshape((beta_size, *beta_shape))
        if effect != effects[0]:  # **** version 0.6 ???
            t = t[1:]
            # fixed effect beta[0] for all additional effect terms, EXCEPT FIRST
            # -> ONE overall theta intercept parameter across all effect terms
        # expand t.ndim to emf.scenario.ndim:
        t = t.reshape(t.shape + tuple(np.ones(len(emf.scenarios) - beta_ndim,
                                              dtype=int)
                                      ))
        ax_0 = range(1, 1 + len(beta_shape))
        ax_new = tuple(list(emf.scenarios).index(e_i) + 1
                       for e_i in effect)
        t = np.moveaxis(t, ax_0, ax_new)
        return t + np.zeros(emf.scenario_shape)
    # ---------------------------------------------------------------
    th = np.concatenate([theta_one(e_tuple)
                         for e_tuple in effects],
                        axis=0)
    return th.reshape((th.shape[0], -1))  # keep as 2D for simplicity


# ------------------------------------------------- TEST:
if __name__ == '__main__':
    from scipy.optimize import approx_fprime, check_grad

    from ema_data import EmaFrame
    from ema_latent import Bradley

    print('*** Testing some ema_base module functions ')
    emf = EmaFrame(scenarios={'CoSS': [f'{i}.' for i in range(1, 8)],
                              'Viktigt': ('Lite',
                                          'Ganska',
                                          'Mycket'),
                              # 'Test': (None, None),
                              'HA': ('A', 'B')
                              },  # nominal variables
                   # stage_key='Stage',
                   attribute_grades={'Speech': ('Svårt',
                                                'Normalt',
                                                'Lätt'),
                                     'Quality': ('Bad', 'Good')
                                     },  # ordinal variables
                   )

    # print('emf=\n', emf)

    main_theta = {'Stage': [0.],
                  'CoSS': 0.1 * np.arange(len(emf.scenarios['CoSS'])),
                  'Viktigt': np.array([-1., 0., 1.]),
                  'HA': np.array([0., 1.])
                  }
    # = only main effects, additive
    regr_effects = ['HA', ('CoSS', 'Viktigt')]

    # effect_HA = np.array([-1., 1.])
    # effect_CoSS = 0.1 * np.arange(len(emf.scenarios['CoSS']))
    # effect_Vikt = np.array([-1., 0., 1.])
    true_theta = main_theta['CoSS'][:, None] + main_theta['Viktigt']
    true_theta = true_theta[..., None] + main_theta['HA']
    true_theta = true_theta[None, ...]  # only one stage
    # stored like emf.scenarios

    beta_HA = main_theta['HA'][1:] - main_theta['HA'][0]
    # beta_CoSS = effect_CoSS[1:] - effect_CoSS[0]
    # beta_Vikt = effect_Vikt[1:] - effect_Vikt[0]
    beta_CoSS_Vikt = main_theta['CoSS'][:, None] + main_theta['Viktigt']
    beta_all = np.concatenate((beta_HA, beta_CoSS_Vikt.reshape((-1)), ))

    eta = [np.zeros(len(r_cat))
           for r_cat in emf.attribute_grades.values()]

    lp_scenarios = np.zeros(emf.scenario_shape).reshape((-1))
    beta_eta = [np.concatenate((beta_all, eta_r))
                for eta_r in eta]
    print(f'len(beta_eta): ', [len(be_i) for be_i in beta_eta])

    xi = np.concatenate((lp_scenarios, *beta_eta))
    xi = xi[None, :]

    # p_base = EmaParamBase()
    p_base = EmaParamBase.initialize(5, emf, regr_effects, rv_class=Bradley)
    print('p_base= ', p_base)
    print(f'p_base.emf=\n{p_base.emf}')
    print(f'p_base.effects= {p_base.effects}')
    print('p_base.scenario_slices', p_base.scenario_slices)
    print('p_base.attribute_slices', p_base.attribute_slices)
    print('p_base.n_parameters= ', p_base.n_parameters)
    print('p_base.n_beta= ', p_base.n_beta)
    print('p_base.theta_map.shape= ', p_base.theta_map.shape)
    print('p_base.comp_prior= ', p_base.comp_prior)

    print('\n*** Testing param extraction methods ***')

    for (stage, alpha_slice) in zip(emf.scenarios[emf.stage_key], p_base.scenario_slices):
        print(f'{emf.stage_key}= {stage}: xi[:, alpha_slice].shape= ', xi[:, alpha_slice].shape)
    # for (stage, alpha) in zip(emf.scenarios[emf.stage_key], p_base.scenario_logprob(xi)):
    #     print(f'{emf.stage_key}= {stage}: prob-mass=\n', np.exp(alpha))
    print(f'\nemf.scenario_shape= {emf.scenario_shape}')
    for (a, a_slice) in zip(emf.attribute_grades.keys(), p_base.attribute_slices):
        print(f'\nAttribute {a}.slice: {a_slice}')
        print(f'\nAttribute {a}: xi[:, a_slice]={xi[:, a_slice]}')

    for a in emf.attribute_grades.keys():
        theta = p_base.attribute_theta(xi, a)
        print(f'\nAttribute {a}: theta.shape={theta.shape}.')
        sc_dim_0 = 'Stage'
        sc_dim_1 = 'CoSS'
        theta = theta.reshape((-1, *emf.scenario_shape))  # *** just for testing
        for (sc_0, theta_0, true_theta_0) in zip(emf.scenarios[sc_dim_0],
                                                 theta[0], true_theta):
            print(f'\t{sc_dim_0}= {sc_0}:')
            for (sc_1, theta_1, true_theta_1) in zip(emf.scenarios[sc_dim_1],
                                                     theta_0, true_theta_0):
                print(f'\t\t{sc_dim_1}= {sc_1}: theta=\n', theta_1)
                print(f'\t\t{sc_dim_1}= {sc_1}: true_theta=\n', true_theta_1)

    for a in emf.attribute_grades.keys():
        tau = p_base.attribute_tau(xi, a)
        print(f'\nAttribute {a}: tau.shape={tau.shape}. tau=\n', tau)

    print('\n*** Testing gradient functions ***')
    print('\n*** Testing p_base._d_tau:')

    eta = np.arange(4)
    limit = 3

    def fun(eta):
        return response_thresholds(eta.reshape((1, -1)))[0, limit]

    def jac(eta):
        return d_response_thresholds(eta.reshape((1, -1)))[0, limit]

    print(f'tau(eta) = {response_thresholds(eta.reshape((1, -1)))}')
    print(f'tau(eta+d) = {response_thresholds(eta.reshape((1, -1)) + .1)} Should remain the same')

    print('approx gradient = ', approx_fprime(eta, fun, epsilon=1e-6))
    print('exact  gradient = ', jac(eta))
    err = check_grad(fun, jac, eta, epsilon=1e-6)
    print('check_grad err = ', err)




"""Run Bayesian analysis of data from an Ecological Momentary Assessment (EMA) study.
This script should be used as a template,
to be copied and modified for any desired analysis.

*** Usage: four main steps, see also explicit template example below

*1: Create an EmaFrame instance to define experiment and select input data.
*2: Load a set of collected EMA data into an EmaDataSet instance
*3: Initialize and train an EmaModel instance corresponding to observed data.
*4: Display results and save figures and tables to a directory tree.


*** Version history:

* Version 0.7.1:
2022-01-30, allow seed input to EmaModel.initialize for reproducible random results

* Version 0.7: minor update to include new calculation and display functions

* Version 0.6:
2021-12-08, allow user control of model restriction:
            restrict_attribute: sensory-variable location average forced -> 0.
            restrict_threshold: response-threshold median forced -> 0.

* Version 0.5:
2021-11-22, first functional version
"""
import numpy as np
from pathlib import Path
import logging
import pickle

from EmaCalc.ema_data import EmaFrame, EmaDataSet
from EmaCalc.ema_model import EmaModel
from EmaCalc.ema_display import EmaDisplaySet
from EmaCalc import ema_logging, __version__

# -------- __main__ check to prevent multiprocessor sub-tasks to re-run this script
if __name__ == '__main__':
    # ------------------------ Set up working directory and result logging:
    work_path = Path.home() / 'Documents' / 'EMA_sim'  # or whatever...
    data_path = work_path / 'data'  # to use simulation data generated by run_sim.py
    result_path = work_path / 'result'  # or whatever

    model_file = 'test_ema_model.pkl'  # name of saved model file (if saved)

    ema_logging.setup(save_path=result_path,
                      log_file='run_ema_log.txt')  # to save the log file

    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    logger.info(f'*** Running EmaCalc version {__version__} ***')

    # ------ 1: Define Experimental Framework: Scenarios, Attributes, and Grades

    # NOTE: This example uses data generated by template script run_sim.py
    # Edit as needed for any other EMA data source

    sim_scenarios = {'Stage': ('',),  # only ONE Test Stage with empty label
                     'HA': ('A', 'B'),  # Two Hearing-aid programs
                     'CoSS': [f'{i}.' for i in range(1, 8)],    # Seven CoSS categories
                     }  # nominal variables, same for all (Sub-)Populations
    # NOTE: First scenario dimension is always Stage, even if only ONE category
    # User may set arbitrary stage_key label
    # Dimension 'Stage' may be omitted, if only one category

    emf = EmaFrame(scenarios=sim_scenarios,
                   stage_key='Stage',
                   attribute_grades={'Speech': ['Very Hard',
                                                'Hard',
                                                'Easy',
                                                'Very Easy',
                                                'Perfect'],
                                     'Comfort': ['Bad',
                                                 'Not Good',
                                                 'Not Bad',
                                                 'Good']},  # ordinal rating categories
                   )

    # ------ 2: Load data from previously saved file(s):

    # NOTE: if 'grouping' is defined, group directory names must match given labels,
    #   as exemplified in module ema_data doc-string
    # With template example: data_path / 'Age_old' is top of directory tree with data files.
    # If several 'grouping' factors are defined, the directory three structure must match the factors.
    # If no grouping is defined: data_path is top of directory tree with all data files

    ds = EmaDataSet.load(emf, data_path, fmt='xlsx',
                         # grouping=None,  # default: include all subjects as ONE unnamed group
                         grouping={'Age': ('old',),  # analyze only Age=old
                                   },
                         # grouping={'Age': ('young','old')},  # analyze both Age groups separately
                         subject='sheet',   # subject ID defined in xlsx sheet title
                         ema_vars={'HA': 'B',
                                   'CoSS': 'C',
                                   'Speech': 'F',
                                   'Comfort': 'G'}
                         )
    logger.info(f'Using data ds=\n{ds}')

    # ------ 3: Learn Analysis Model from loaded data set:

    # Model ordinal-regression effects of Scenarios on each Attribute:

    # regression_effects = ['HA',     # main linear regression effect only
    #                       'CoSS',   # main linear regression effect only
    #                       # 'Stage',  # if there are several Stage categories
    #                       ]

    regression_effects = [('HA', 'CoSS')  # joint effects, main AND interaction
                          # 'Stage',  # if there are several Stage categories
                          ]

    # NOTE: A regression_effects element may include any combination of scenario dimensions, but
    # including ALL interactions -> many model parameters,
    # possibly -> less precise estimation for each parameter.

    # In this example: ['HA', 'CoSS'] -> 2 + (7 - 1) = 8 regression-effect parameters
    #                ['CoSS', 'HA'] -> 7 + (2 - 1) = 8 regression-effect parameters
    #                [('HA', 'CoSS')] -> 2 * 7 = 14 regression-effect parameters

    emm = EmaModel.initialize(ds,
                              effects=regression_effects,
                              max_n_comp=10,
                              restrict_attribute=False,  # default
                              restrict_threshold=True,  # default
                              # seed=12345  # ONLY if reproducible results are required
                              )
    # max_n_comp = max number of mixture components in population model
    # restrict_attribute=True -> force sensory-variable locations centered at zero
    # restrict_threshold=True -> force middle response threshold(s) centered at zero
    # for each respondent and each sample of each attribute

    ll = emm.learn(max_hours=2., max_minutes=0.)
    logger.info('ll= ' + np.array2string(np.array(ll), precision=5))
    emm.prune()  # keep only active mixture components.
    # NOTE: If the number of components is not reduced,
    # it might be a good idea to re-run with larger initial max_n_comp,
    # BUT max_n_comp always <= half number of subjects

    # -------- Save learned EmaModel (optional):
    with (work_path / model_file).open('wb') as f:
        pickle.dump(emm, f)

    # ------ 4: Generate result displays:

    emd = EmaDisplaySet.show(emm,
                             scenarios=['CoSS',  # CoSS profile, aggregated across HA
                                        ('CoSS', 'HA'),  # CoSS profiles, conditional on HA
                                        ('HA', 'CoSS'),  # HA profiles, conditional on CoSS
                                        ],
                             attributes=[('Speech', 'CoSS'),  # Speech, main effect of CoSS
                                         ('Speech', 'HA'),  # Speech, main effect of HA
                                         ('Speech', ('CoSS', 'HA')),  # joint effect of both
                                         ('Comfort', ('CoSS', 'HA'))],  # joint effect of both
                             population_individual=True,  # random individual in population
                             population_mean=True,  # population mean
                             subjects=False,  # individual results: True -> MANY plots and tables
                             nap=[('Speech', 'HA'),  # NAP(HA) grades aggregated across CoSS
                                  ('Speech', ('HA', 'CoSS')),  # NAP(HA) separated by CoSS
                                  ('Comfort', 'HA')],
                             nap_confidence=0.95,  # level of symmetric confidence interval
                             grade_thresholds=True,  # response thresholds in attribute plots
                             percentiles=[2.5, 25, 50, 75, 97.5],  # in profile plots and tables
                             credibility_limit=0.7,  # minimum credibility in difference tables
                             figure_format='pdf',  # or 'eps', 'jpg', 'png'
                             table_format='latex',  # or 'tab' for tab-limited text
                             )
    # NOTE: joint (=interaction) effects are correct only if included in model regression_effects
    # nap results calculated only for Scenario dimension with exactly TWO categories.

    # ------------------------------- save all result displays
    emd.save(result_path)

    logging.info(f'All results saved in {result_path}')

    logging.shutdown()



"""Superclasses for input data files.
"""


# ------------------------------------------------ File exceptions
class FileReadError(RuntimeError):
    """Any type of exception during file read
    to be sub-classed for more specific variants depending on file format
    """


# -----------------------------------------------------------------------
class EmaRecord:  # **** might be just a dict, incl key 'subject' ?
    """Container for a single EMA record
    """
    def __init__(self,
                 subject=None,
                 ema=None
                 ):
        """
        :param subject: string or any other immutable type
        :param ema: tuple of one or more tuples (cf, c_category)
        """
        self.subject = subject
        self.ema = ema

    def __repr__(self):
        return (self.__class__.__name__ + '(\n\t' +
                ',\n\t'.join(f'{key}={repr(v)}'
                             for (key, v) in vars(self).items()) +
                '\n\t)')


# ---------------------------------------------------------
def safe_file_path(p):
    """Ensure non-existing file path, to avoid over-writing,
    by adding a sequence number to the path stem
    :param p: file path
    :return: p_safe = similar file path with modified name
    """
    f_stem = p.stem
    f_suffix = p.suffix
    i = 0
    while p.exists():
        i += 1
        p = p.with_name(f_stem + f'-{i}.' + f_suffix)
    return p


# ---------------------------------------------------------
class EmaFile:
    """Superclass for any container of EMA records
    from one or more Subjects, participating in an EMA study with
    one or more nominal Scenarios, and zero or more ordinal Attributes,
    accessed by an iterator yielding EmaRecord instances.

    This super-class is used mainly for READING a data file.

    Sub-classes must be defined for each specific file format.
    Sub-classes MUST define an __iter__ method, yielding EmaRecord instances
    """
    def __init__(self, file_path):
        """
        :param file_path: Path instance for READING, or None.
            Another path may be defined in the save method.
        """
        self.file_path = file_path

    def __repr__(self):
        return (self.__class__.__name__ + '(\n\t' +
                ',\n\t'.join(f'{key}={repr(v)}'
                             for (key, v) in vars(self).items()) +
                '\n\t)')

    def __iter__(self):
        raise NotImplementedError




"""This module defines classes and methods for simulated EMA study
with groups of respondents drawn at random from population(s)
with specified inter-individual distributions of all model parameters.

The simulations can generate artificial data with the same structure
as a real experiment.


*** Main Module Classes:

EmaSimPopulation: defines distribution of model parameters
    for probabilities of one or more SCENARIOS, in one or more Scenario Dimensions,
    and one or more perceptual ATTRIBUTE latent-variable locations
    in ONE (sub-)population of subjects.

EmaSimExperiment: defines a complete EMA experiment,
    generates simulated responses by subjects in one or more groups,
    sampled from EmaSimPopulation instance(s).
    The EMA data layout is defined by an ema_data.EmaFrame instance,
    defining all Scenario dimensions and categories, and
    defining discrete ordinal response scales for each Attribute.

EmaSimSubject = superclass for subject properties

SubjectBradley: single subject with Attribute ratings determined by
    the Bradley-Terry-Luce (BTL) model, assuming standard Logistic latent sensory variable,
    i.e., with st.dev. approx= 1.8

SubjectThurstone: single subject with Attribute ratings determined by
    the Thurstone Case V model, assuming standard Gaussian latent sensory variable,
    i.e., with st.dev. = 1.

EmaSimGroup: container for EmaSimSubject instances
    drawn at random from an EmaSimPopulation instance.


*** Main Class Methods:

EmaSimPopulation.gen_group(...)
    draws a group of simulated subjects at random from the simulated Population.

EmaSimExperiment.gen_dataset(...)
    generates am ema_data.EmaDataSet instance with simulated EMA records
    for one or more groups of simulated subjects.
    All records can be saved to files using the EmaDataSet.save(...) method.

*** Usage example: See script run_sim.py

*** Version History:

* Version 0.7.1:
2022-01-19, module function set_sim_seed to ensure reproducibility
2022-01-19, EmaSimSubject.rvs methods defined here, to ensure reproducibility
2022-01-06, EmaSimPopulation.response_width_mean to control individual decision thresholds
2022-01-05, minor cleanup EmaSimExperiment

* Version 0.5.1:
2021-11-25, allow experiment with NO Attributes

* Version 0.5:
2021-11-14, copied and modified PairedCompCalc -> EmaCalc
2021-11-16, first functional version
2021-11-21, rather extensively tested
"""
# **** specify separate Attribute variance for each Scenario dimension ????
# **** separate inter-individual attribute variance jointly for all scenarios,
# **** plus variance for independent variations across scenarios???

import numpy as np
import logging

from .ema_data import EmaDataSet
from .ema_base import response_thresholds, tau_inv
from . import ema_latent


RNG = np.random.default_rng()
# = default module-global random generator
# may be modified via function set_sim_seed

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # test


# ------------------ module function
def set_sim_seed(seed=None):
    """Set module-global RNG with given seed.
    To be called by user BEFORE any other work,
    to achieve reproducible results.
    :param seed: (optional) integer
    :return: None
    """
    global RNG
    RNG = np.random.default_rng(seed)
    if seed is not None:
        logger.warning(f'*** Using seed={seed} -> reproducible results.')


# --------------------------------------- subject response models
class EmaSimSubject:
    """Simulate one individual participant in an EMA data-collection experiment.
    Superclass for either SubjectBradley or SubjectThurstone
    """
    def __init__(self, sc_prob, a_theta, a_tau):  # lapse_prob=0.):
        """
        :param sc_prob: mD array with (non-normalized) conditional Scenario probability
            sc_prob[k0, k1, k2,...] propto Prob (k1, k2, ...)-th scenario, GIVEN k0-th Stage
            sc_prob.shape == emf.scenario_shape
        :param a_theta: dict with (a_key, a_theta) elements, where
            a_key = a string identifying ONE Attribute among emf.attribute_grades.keys()
            a_theta[k0, k1, ...] = mD array with mean of latent sensory variable for Attribute a_key,
        :param a_tau: dict with (a_key, thr) elements, where
            tnr[l] = UPPER interval limit for l-th ordinal response,
            NOT INCLUDING extreme -inf, +inf limits
        """
        self.sc_prob = sc_prob
        self.a_theta = a_theta
        self.a_tau = a_tau

    def __repr__(self):
        return (f'{self.__class__.__name__}(\n\t' +
                ',\n\t'.join(f'{key}={repr(v)}'
                             for (key, v) in vars(self).items()) +
                '\n\t)')

    def gen_ema_records(self, emf, min_ema, max_ema):
        """Generate a sequence of EMA records at random
        using response properties of self.
        :param emf: ema_data.EmaFrame instance
        :param min_ema: min random number of EMA records per Stage
        :param max_ema: max random number of EMA records per Stage
        :return: ema_list = sequence of dicts with elements (key, cat), where
            key is one of emf.scenarios or emf.attribute_grades, and
            cat is one of the allowed categories for that key
            len(ema_list) == n_records randomly drawn with
            min_ema <= n_records < max_ema
        """
        def scenario_index(sc_p, sc_shape):
            """Generate ONE random scenario index tuple, NOT including Stage index
            :param sc_p: 1D probability-mass array for all scenarios, EXCEPT Stage
            :param sc_shape: tuple with scenario_shape EXCEPT Stage dimension
            :return: ind = index tuple; len(ind) = len(sc_shape)
             """
            sc_ind = RNG.choice(len(sc_p), p=sc_p)  # linear random index
            return np.unravel_index(sc_ind, shape=sc_shape)

        def scenario_dict(sc_ind):
            """Convert index tuple to Scenario dict
            """
            return {sc_k: sc_cat[i]
                    for (i, (sc_k, sc_cat)) in zip(sc_ind,
                                                   emf.scenarios.items())}
        # -------------------------------------------------------

        ema_list = []
        sc_shape_stage = emf.scenario_shape[1:]
        sc_prob_stage = self.sc_prob.reshape((self.sc_prob.shape[0], -1))
        for (i, p) in enumerate(sc_prob_stage):
            n_rec = RNG.integers(min_ema, max_ema)
            for _ in range(n_rec):
                sc = (i, *scenario_index(p, sc_shape_stage))
                a_grades = self.gen_attr_grades(emf, sc)
                ema_list.append(scenario_dict(sc) | a_grades)
        return ema_list

    def gen_attr_grades(self, emf, sc):
        """Generate random ordinal Attribute grades in given Scenario
        :param emf: external ema_data.EmaFrame object defining experimental layout
        :param sc: index tuple for scenario of this record
        :return: a_grades = dict with elements (a, grade)
        """
        def rvs_grade(th, tau):
            x = self.rvs(th)  # done by sub-class
            return np.sum(x > tau)
        # -------------------------------------------------------
        return {a: emf.attribute_grades[a][rvs_grade(a_th[sc], self.a_tau[a])]
                for (a, a_th) in self.a_theta.items()}

    @staticmethod
    def rvs(th):
        """Abstract method, implemented by sub-class
        """
        raise NotImplementedError

    # def lapse(self):  # *** future ???
    #     """Generate True with prob = self.lapse_prob
    #     """
    #     return uniform.rvs(0, 1.) < self.lapse_prob
    #
    # def lapse_response(self):
    #     """Generate a random result, disregarding quality parameters
    #     :return: scalar integer
    #         in {-n_difference_grades, ...,-1, +1,..., + n_difference_grades}, if forced_choice
    #         in {-n_difference_grades+1, ...,0, ...,  + n_difference_grades-1}, if not forced_choice
    #         i.e., excluding 0 if self.emf.forced_choice
    #     """
    #     n_response_limits = len(self.response_thresholds)
    #     # if self.emf.forced_choice:
    #     if self.response_thresholds[0] == 0.:  # forced_choice
    #         return ((-1 if uniform.rvs() < 0.5 else +1) *
    #                 randint.rvs(low=1, high=n_response_limits + 1))
    #
    #     else:
    #         return randint.rvs(low=-n_response_limits,
    #                            high=n_response_limits + 1)
    #


class SubjectThurstone(ema_latent.Thurstone, EmaSimSubject):
    """Simulate a subject using the Thurstone Case V choice model.
    """
    @staticmethod
    def rvs(loc, size=None):
        """Draw random variables from self
        :param loc: scalar or array-like location
        :param size: (optional) size of result
        :return: x = scalar or array
            x.shape == loc.shape, if no size is given
        """
        loc = np.asarray(loc)
        if size is None:
            size = loc.shape
        return loc + RNG.standard_normal(size=size)


class SubjectBradley(ema_latent.Bradley, EmaSimSubject):
    """Simulate one individual participant in a paired-comparison experiment.
    The subject responds using the Bradley-Terry-Luce choice model,
    with parameters defined in the log domain.
    """
    @staticmethod
    def rvs(loc, size=None):
        """Draw random variables from self
        :param loc: scalar or array-like location
        :param size: (optional) size of result
        :return: x = scalar or array
            x.shape == loc.shape
        """
        return RNG.logistic(loc, size=size)


# -------------------------------------------------------------------------
class EmaSimPopulation:
    """Defines a simulated population
    from which groups of random test subjects can be generated
    for a simulated EMA experiment.

    The population instance defines distributions for
    one or more nominal Scenario categories, and
    zero, one or more perceptual Attributes, given any Scenario.

    Each Scenario is a combination of one category from each Scenario Dimension.
    """
    def __init__(self, emf,
                 scenario_prob,
                 attribute_mean=None,
                 log_scenario_std=0.,
                 attribute_std=0.,
                 response_width_mean=None,  # -> subject_class.scale
                 log_response_width_std=0.,
                 lapse_prob_range=(0., 0.),  # ******* not needed, not used ******
                 subject_class=SubjectBradley,
                 id=''):
        """
        :param emf: ema_data.EmaFrame instance defining Scenarios and Attributes
        :param scenario_prob: array-like multi-dim, with
            scenario_prob[k0, k1,...] prop.to probability of (k1, k2,...)-th scenario,
            GIVEN the k0-th TestStage category, even if only one stage category.
            scenario_prob.shape must correspond to shape of scenarios.
        :param attribute_mean: (optional) dict or iterable with elements (a_key, a_mean), where
            a_key is string id of a rated perceptual attribute,
            a_mean is an mD array with
            a_mean[k0, k1, ...] = latent-variable mean for attribute a_key,
            given the (k0, k1,...)-th Scenario category, i.e.,
            a_mean.shape == scenario_prob.shape, for all attributes.
        :param log_scenario_std: (optional) inter-individual standard deviation of
            log probabilities for each scenario category.
        :param attribute_std: (optional) scalar inter-individual standard deviation of all attribute parameters
        :param response_width_mean: (optional) mean width of response intervals,
            in subject_class scale units
        :param log_response_width_std: (optional) scalar standard deviation of log(response-interval-width)
        :param lapse_prob_range: (optional) tuple (min, max) probability of random lapse response
        :param subject_class: (optional) subject probabilistic model for generating responses
        :param id: (optional) string label, used as prefix in all generated subject names
        """
        self.emf = emf  # save emf ref here, too, although same all sub-populations
        scenario_prob = np.asarray(scenario_prob)
        if emf.scenario_shape[0] == 1 and emf.scenario_shape[1:] == scenario_prob.shape:
            scenario_prob = scenario_prob[None, ...]
        if emf.scenario_shape != scenario_prob.shape:
            raise RuntimeError('scenario_prob.shape must agree with EmaFrame scenarios')
        self.scenario_prob = scenario_prob
        for (i, sc_i) in enumerate(self.scenario_prob):
            self.scenario_prob[i] /= np.sum(sc_i)  # normalize conditional probabilities, given Stage
        self.log_scenario_std = log_scenario_std

        if attribute_mean is None:
            attribute_mean = dict()  # NO attributes
        else:
            attribute_mean = dict(attribute_mean)
        if set(attribute_mean.keys()) != set(emf.attribute_grades.keys()):
            raise RuntimeError('attribute_mean must define same attributes as EmaFrame')
        for (a, a_mean) in attribute_mean.items():
            if (emf.scenario_shape[0] == 1
                    and emf.scenario_shape[1:] == a_mean.shape):
                attribute_mean[a] = attribute_mean[a][None, ...]
        assert all(a_mean.shape == emf.scenario_shape
                   for a_mean in attribute_mean.values()), 'attribute_mean.shape must match scenario_shape'
        self.attribute_mean = attribute_mean
        self.attribute_std = attribute_std
        if response_width_mean is None:
            response_width_mean = subject_class.scale
        self.response_width_mean = response_width_mean
        self.log_response_width_std = log_response_width_std
        self.lapse_prob_range = lapse_prob_range
        self.subject_class = subject_class
        self.id = id

    def __repr__(self):
        return (f'{self.__class__.__name__}(\n\t' +
                ',\n\t'.join(f'{key}={repr(v)}'
                             for (key, v) in vars(self).items()) +
                '\n\t)')

    @property
    def n_attributes(self):
        return len(self.attribute_mean)

    def gen_group(self, n_subjects=1):
        """Create a group of simulated-subject instances randomly drawn from self,
        with properties suitable for a planned experiment.
        :param n_subjects: number of randomly drawn subjects from self
        :return: single EmaSimGroup instance, containing generated subjects,
            each with properties drawn from self.
        """
        def gen_sc_prob():
            sc_p = np.exp(np.log(self.scenario_prob) +
                          self.log_scenario_std *
                          RNG.standard_normal(size=self.emf.scenario_shape))
            for (i, sc_i) in enumerate(sc_p):
                sc_p[i] /= np.sum(sc_i)  # normalize conditional probabilities, given Stage
            return sc_p

        def gen_attr():
            theta = self.attribute_std * RNG.standard_normal(size=(self.n_attributes,
                                                                   *self.emf.scenario_shape))
            # theta[i, ...] = random offset of location for i-th attribute of s-th subject
            return {a: th_mean + th_d
                    for ((a, th_mean), th_d) in zip(self.attribute_mean.items(),
                                                    theta)}

        def gen_tau():
            """Random threshold parameters for each attribute
            """
            def tau(n_levels):
                """Calc response thresholds
                :param n_levels: integer number of response intervals
                :return: t = 1D array with INNER interval limits, i.e.,
                    t[m] = UPPER interval limit for m-th ordinal response
                    t.shape == (n_levels - 1,)
                """
                t = self.response_width_mean * (np.arange(n_levels - 1) + 1. - n_levels / 2)
                log_w = tau_inv(t) + self.log_response_width_std * RNG.standard_normal(size=n_levels)
                # incl. random variations of log interval width in mapped [0, 1] range
                return response_thresholds(log_w)[1:-1]  # EXCL (-inf, +inf) extremes
            # ----------------------------------------------------
            return {a: tau(len(a_grades))
                    for (a, a_grades) in self.emf.attribute_grades.items()}
        # --------------------------------------------------------------------

        return EmaSimGroup(self,
                           {self.id + f'_S{i}': self.subject_class(sc_prob=gen_sc_prob(),
                                                                   a_theta=gen_attr(),
                                                                   a_tau=gen_tau())
                            for i in range(n_subjects)})


# ----------------------------------------------------------------------
class EmaSimGroup:
    """Group of test subjects drawn from a given population
    """
    def __init__(self, pop, subjects):
        """
        :param pop: an EmaSimPopulation instance,
            from which subjects have been drawn at random
        :param subjects: dict with (s_id, s_sim)
            s_id = string id for the subject
            s_sim = an EmaSimSubject instance
        """
        self.pop = pop
        self.subjects = subjects

    def __repr__(self):
        return (self.__class__.__name__ + '('
                + f'\n\tpop={repr(self.pop)}'
                + f'\n\tsubjects= dict with {len(self.subjects)} simulated subjects)')


# -------------------------------------------------------------------------
class EmaSimExperiment:
    """Defines a simulated EMA data-collection experiment,
    with one or more groups of simulated subjects, with
    each group generated from an EmaSimPopulation instance.

    Method gen_dataset() generates a complete ema_data.EmaDataSet instance
    with EMA records for all subjects in all groups.

    The experimental procedure is defined by
    emf = an ema_data.EmaFrame instance
    """
    def __init__(self, emf, groups):
        """
        :param emf: EmaFrame instance, defining experimental parameters
        :param groups: dict with elements (g_id, g_sim),
            g_sim = an EmaSimGroup instance
        """
        self.emf = emf
        self.groups = groups

    def __repr__(self):
        return (f'{self.__class__.__name__}(\n\t' +
                ',\n\t'.join(f'{key}={repr(v)}'
                             for (key, v) in vars(self).items()) +
                '\n\t)')

    def gen_dataset(self, min_ema=3, max_ema=50):
        """Generate a complete EmaDataSet instance for this experiment,
        with one or more groups of subjects.
        :param min_ema: min random number of EMA records in each Stage
        :param max_ema: max random number of EMA records in each Stage
        :return: a single EmaDataSet instance
        """
        emd = {g: {s_id: s.gen_ema_records(self.emf, min_ema, max_ema)
                   for (s_id, s) in g_sim.subjects.items()}
               for (g, g_sim) in self.groups.items()}
        return EmaDataSet(self.emf, emd)


# ------------------------------------- internal help function:
# def thurstone2bradley(x):
#     """transform parameters defined on Thurstone d-prime scale,
#     to equivalent parameters on the BTL model scale
#     :param: x = array or array-like list of values in d-prime units
#     :return: array with transformed values
#     """
#     return logistic.ppf(norm(scale=np.sqrt(2)).cdf(x))
#
#
# def bradley2thurstone(x):
#     """transform parameters defined on Bradley scale,
#     to equivalent parameters on the Thurstone model scale
#     :param: x = array or array-like list of values in BTL units
#     :return: array with transformed values
#     """
#     return norm.ppf(logistic.cdf(x), scale=np.sqrt(2))
